#!/usr/bin/env ruby

require 'artifactory'
require 'fileutils'
require 'json'
require 'omnibus'
require 'tempfile'

def self.env_or_raise(key)
  ENV[key] || raise("Required ENV variable `#{key}` is unset!")
end

def self.env_or_empty(key)
  ENV[key] || ''
end

# Values passed into the command sub-process by Jenkins
build_number           = env_or_raise('BUILD_NUMBER')
jenkins_home           = env_or_raise('JENKINS_HOME')
jenkins_workspace      = env_or_raise('WORKSPACE')
pipeline_name          = env_or_raise('PIPELINE_NAME')
project_name           = env_or_raise('PROJECT_NAME')
omnibus_config_path = File.join(jenkins_home, 'omnibus-publish.rb')

Dir.chdir(jenkins_workspace) do
  Omnibus.load_configuration(omnibus_config_path)

  client = Artifactory::Client.new(
    endpoint:       Omnibus::Config.artifactory_endpoint,
    username:       Omnibus::Config.artifactory_username,
    password:       Omnibus::Config.artifactory_password,
    ssl_pem_file:   Omnibus::Config.artifactory_ssl_pem_file,
    ssl_verify:     Omnibus::Config.artifactory_ssl_verify,
    proxy_username: Omnibus::Config.artifactory_proxy_username,
    proxy_password: Omnibus::Config.artifactory_proxy_password,
    proxy_address:  Omnibus::Config.artifactory_proxy_address,
    proxy_port:     Omnibus::Config.artifactory_proxy_port,
  )

  # Get the build version
  glob_pattern_base = "#{jenkins_home}/jobs/#{pipeline_name}-build/configurations/axis-architecture/*/axis-platform/*/axis-project/#{project_name}/axis-role/builder/builds/#{build_number}/archive"
  dirglob = Dir.glob("#{glob_pattern_base}/**/*.metadata.json").first
  metadata = JSON.parse(File.read(dirglob), symbolize_names: true)
  build_version = metadata[:version]
  build_record_name = "#{project_name}##{build_version}"

  # Check whether the build record exists
  begin
    response = client.get("/api/build/#{project_name}/#{build_version}")
  rescue Artifactory::Error::HTTPError => e
    raise e unless e.message.include?('No build was found')
  end

  raise "A build record already exists for #{build_record_name}!" unless response.nil?

  build_record_properties = {
    'omnibus.project' => project_name,
    'omnibus.version' => build_version,
  }

  params = build_record_properties.map { |k, v| "#{k}=#{v}" }.join('|')
  path   = ['/api/plugins/execute/create_build', "params=#{params}"].compact.join('?')
  path = URI.encode(path)
  url = path.gsub('+', '%2B')

  begin
    client.post(url, nil)

    puts "Successfully created build record in artifactory for #{build_record_name}"
  rescue ::Artifactory::Error::HTTPError => e
    raise <<-EOH.gsub(/^ {6}/, '')
      There was an error creating a build record in artifactory for #{build_record_name}!

      Full error message from #{Omnibus::Config.artifactory_endpoint}:

      #{e.message}
     EOH
  end
end
