#!/usr/bin/env ruby

require 'artifactory'
require 'omnibus'
require 'mixlib/versioning'

def self.env_or_raise(key)
  ENV[key] || raise("Required ENV variable `#{key}` is unset!")
end

def self.omnibus_arch(platform, _platform_version, jenkins_arch)
  case platform
  when 'debian', 'el', 'ubuntu'
    if jenkins_arch == 'i386'
      'i686'
    else
      jenkins_arch
    end
  when 'freebsd'
    if jenkins_arch == 'x86_64'
      'amd64'
    else
      jenkins_arch
    end
  when 'solaris'
    if jenkins_arch == 'i86pc'
      'i386'
    elsif %w(sun4u sun4v).include?(jenkins_arch)
      'sparc'
    end
  else
    jenkins_arch
  end
end

def self.omnibus_platform(platform, _platform_version)
  case platform
  when 'solaris'
    'solaris2'
  else
    platform
  end
end

def self.omnibus_platform_version(platform, platform_version)
  case platform
  when 'solaris'
    "5.#{platform_version}"
  else
    platform_version
  end
end

# Values passed into the command sub-process by Jenkins
jenkins_home      = env_or_raise('JENKINS_HOME')
jenkins_workspace = env_or_raise('WORKSPACE')
project_name      = env_or_raise('PROJECT_NAME')
architecture      = env_or_raise('architecture')
test_platform     = env_or_raise('TEST_PLATFORM')

# We use 'acceptance' as a platform name to add our acceptance testers
# into the build / test matrix. Since we can not use 'acceptance' as a
# platform name, we translate it to the platform name of our acceptance
# testers. If we add different types of acceptance testers in the future
# we will need to revisit this line.
if test_platform == 'acceptance'
  # chefdk, chef, and angrychef all build on ubuntu-14.04
  test_platform = 'ubuntu-14.04'
end

build_version   = ENV['OMNIBUS_BUILD_VERSION']
delivery_change = ENV['DELIVERY_CHANGE']

omnibus_config_path = File.join(jenkins_home, 'omnibus-publish.rb')

platform = test_platform.split('-')[0]
platform_version = test_platform.split('-')[1]

channel = 'omnibus-unstable-local'
channel = 'omnibus-current-local' if delivery_change.nil? && build_version.nil?

Dir.chdir(jenkins_workspace) do
  Omnibus.load_configuration(omnibus_config_path)

  Artifactory.configure do |config|
    config.endpoint       = Omnibus::Config.artifactory_endpoint
    config.username       = Omnibus::Config.artifactory_username
    config.password       = Omnibus::Config.artifactory_password
    config.ssl_pem_file   = Omnibus::Config.artifactory_ssl_pem_file
    config.ssl_verify     = Omnibus::Config.artifactory_ssl_verify
    config.proxy_username = Omnibus::Config.artifactory_proxy_username
    config.proxy_password = Omnibus::Config.artifactory_proxy_password
    config.proxy_address  = Omnibus::Config.artifactory_proxy_address
    config.proxy_port     = Omnibus::Config.artifactory_proxy_port
  end

  properties = {
    'repos' => channel,
    'omnibus.project' => project_name,
    'omnibus.architecture' => omnibus_arch(platform, platform_version, architecture),
    'omnibus.platform' => omnibus_platform(platform, platform_version),
    'omnibus.platform_version' => omnibus_platform_version(platform, platform_version),
  }

  if build_version
    properties['build.number'] = build_version
  elsif delivery_change
    properties['delivery.change'] = delivery_change
  else
    puts "Finding the latest version of #{project_name} since neither 'OMNIBUS_BUILD_VERSION' nor 'DELIVERY_CHANGE' is set!"

    query = <<-QUERY
items.find(
  { "repo": "#{channel}" },
  { "@omnibus.project": "#{project_name}" }
).include("repo", "path", "name", "property")
    QUERY

    results = Artifactory.post('/api/search/aql', query, 'Content-Type' => 'text/plain')
    versions = results['results'].collect { |r| r['properties'] }
                                 .flatten
                                 .find_all { |elt| elt['key'] == 'build.number' }
                                 .collect { |elt| elt['value'] }
    build_version = Mixlib::Versioning.find_target_version(versions, nil, false, true).to_s

    properties['build.number'] = build_version
  end

  artifacts = Artifactory::Resource::Artifact.property_search(properties).reject { |artifact| artifact.uri.end_with? '.metadata.json' }
  if artifacts.empty?
    raise "Could not locate artifact #{project_name}##{(build_version || delivery_change)} in #{channel} for #{test_platform} on #{architecture}"
  else
    puts "Downloading artifact #{project_name}##{(build_version || delivery_change)} from #{channel} for #{test_platform} on #{architecture}"
    # Download all the artifacts in parallel to speed up the download process.
    threads = []
    artifacts.each do |artifact|
      #
      # Here we have to sub out the artifactory download uri because it gets
      # passed back from the Artifactory API query.
      #
      if architecture == 's390x'
        artifact.download_uri.sub!('http://artifactory.chef.co', Artifactory.endpoint)
      end
      dl = Thread.new { artifact.download(jenkins_workspace) }
      dl.abort_on_exception = true
      threads << dl
    end
    # Block on downloads so we can finish them before the script terminates.
    threads.each(&:join)
  end
end
