#!/usr/bin/env ruby

require 'artifactory'
require 'fileutils'
require 'json'
require 'omnibus'
require 'tempfile'
require 'rubygems/commands/push_command'

OMNIBUS_PACKAGE_PATTERN = '**/pkg/*.{bff,deb,dmg,msi,p5p,rpm,solaris,amd64.sh,i386.sh}'.freeze

def self.env_or_empty(key)
  ENV[key] || ''
end

def self.env_or_raise(key)
  ENV[key] || raise("Required ENV variable `#{key}` is unset!")
end

def self.fix_metadata(package_local_path)
  package = Omnibus::Package.new(package_local_path)

  old_metadata = JSON.parse(File.read(package.metadata.path), symbolize_names: true)
  new_metadata = old_metadata.dup

  # OLD Omnibus does not include a project name!!!! O_o
  unless new_metadata[:name]
    match = package.name.match(/^(?<project_name>[a-z].*)(_|-)\d\.\d\.\d/)
    new_metadata[:name] = match[:project_name]
  end

  # Historically we name our Windows-specific project definitions
  # PROJECT_NAME-windows. We want to publish this project under
  # PROJECT_NAME though. Luckily we are moving to a place where the
  # Windows-specific project definition will just become part of the
  # regular *nix project definition. Until then we need to modify the
  # the artifact's associated `*.metadata.json` file as the Omnibus
  # publisher uses this metadata when publishing to Artifactory.
  match = new_metadata[:name].match(/^(?<project_basename>.*)-windows$/)
  new_metadata[:name] = match[:project_basename] if match

  # In pre-4.0 Omnibus packages are signed outside of the Omnibus
  # build process which means Omnibus's generated checksums are wrong.
  # As we use the Omnibus-generated metadata when publishing to
  # Artifactory checksum mismatches are OK until Omnibus 4.0 is in
  # general use.
  if new_metadata[:platform] == 'el'
    new_metadata[:md5]    = package.md5
    new_metadata[:sha1]   = package.sha1
    new_metadata[:sha256] = package.sha256
    new_metadata[:sha512] = package.sha512
  end

  # OLD Omnibus does not generate all checksum types
  new_metadata[:sha256] = package.sha256 unless new_metadata[:sha256]
  new_metadata[:sha512] = package.sha512 unless new_metadata[:sha512]

  # If the metadata has changed write the file
  if new_metadata != old_metadata
    metadata = if Omnibus.const_defined?('Metadata')
                 Omnibus::Metadata.new(package, new_metadata)
               else
                 Omnibus::Package::Metadata.new(package, new_metadata)
               end
    puts "\nMetadata has changed...updating '#{metadata.path}'"
    metadata.save
  end
end

# Values passed into the command sub-process by Jenkins
jenkins_home           = env_or_raise('JENKINS_HOME')
jenkins_workspace      = env_or_raise('WORKSPACE')
build_number           = env_or_raise('BUILD_NUMBER')
project_name           = env_or_raise('PROJECT_NAME')
pipeline_name          = env_or_raise('PIPELINE_NAME')
node_name              = env_or_raise('NODE_NAME')
delivery_change        = env_or_empty('DELIVERY_CHANGE')
delivery_sha           = env_or_empty('DELIVERY_SHA')

omnibus_config_path    = File.join(jenkins_home, 'omnibus-publish.rb')

# We'll grab artifacts from the archive location if executing on the master
glob_pattern_base = if node_name == 'master'
                      "#{jenkins_home}/jobs/#{pipeline_name}-build/configurations/axis-architecture/*/axis-platform/*/axis-project/#{project_name}/axis-role/builder/builds/#{build_number}/archive"
                    elsif node_name.include?('windows')
                      jenkins_workspace.tr('\\', '/')
                    else
                      jenkins_workspace
                    end

Dir.chdir(jenkins_workspace) do
  package_glob_pattern = "#{glob_pattern_base}/#{OMNIBUS_PACKAGE_PATTERN}"

  puts "Publishing with glob pattern of #{package_glob_pattern}"
  puts ''

  # Apply various 'fixes' to the metadata before uploading
  Dir.glob(package_glob_pattern).each { |package_path| fix_metadata(package_path) }

  # This test -> build platform mapping file is generated by the opscode-ci cookbook code.
  build_to_test_mappings = JSON.parse(File.read("opscode-ci/platform-mappings/#{pipeline_name}.json"))['build_to_test']

  # We remove acceptance test entry from the build_to_test_mappings
  # in order not to publish any artifacts for "acceptance" platform
  # because it does not exist as a platform.
  build_to_test_mappings.each { |_k, v| v.delete('acceptance') }

  build_to_test_mappings = build_to_test_mappings.map do |k, v|
    # We name the platform/version slightly different in Solaris:
    if k =~ /solaris-(\d)*/
      platform_splitter = proc do |p|
        p, pv = p.rpartition('-') - %w( - )
        "#{p}2-5.#{pv}"
      end

      [platform_splitter.call(k), v.map { |p| platform_splitter.call(p) }]
    else
      [k, v]
    end
  end.to_h

  Omnibus.load_configuration(omnibus_config_path)
  publisher = Omnibus::ArtifactoryPublisher.new(
    package_glob_pattern,
                repository: 'omnibus-unstable-local',
                platform_mappings: build_to_test_mappings,
                build_record: false,
                properties: {
                  'delivery.change' => delivery_change,
                  'delivery.sha' => delivery_sha,
                },
  )

  if publisher.packages.empty?
    raise "Could not locate any #{project_name} artifacts to publish."
  else
    publisher.publish do |package|
      puts "Published '#{package.name}' for #{package.metadata[:platform]}-#{package.metadata[:platform_version]}"
    end

    puts <<-EOH

DONE! \\m/

    EOH
  end

  # We need to push the gems in certain cases (ie chef, chef-dk)
  if %w(chef chefdk inspec).include?(project_name) && (ENV['ADHOC_BUILD'] != 'true')
    GEM_PACKAGE_PATTERN = '**/[^/]*\.gem'.freeze
    gem_base_name = (project_name == 'chefdk') ? 'chef-dk' : project_name
    project_source = "/var/cache/omnibus/src/#{gem_base_name}"
    # This will exclude any gems in a /spec/ directory
    gems_to_publish = Dir.glob("#{project_source}/#{GEM_PACKAGE_PATTERN}") - Dir.glob("#{project_source}/spec/#{GEM_PACKAGE_PATTERN}")

    puts "Publishing Gems from #{project_source}"
    puts ''

    gems_to_publish.each do |gem_path|
      puts 'Publishing gem ' + gem_path
      artifactory_endpoint = "#{Omnibus::Config.artifactory_endpoint}/api/gems/omnibus-gems-local"
      # This mimics the behavior of the gem command line, and is a public api:
      # http://docs.seattlerb.org/rubygems/Gem/Command.html
      gem_pusher = Gem::Commands::PushCommand.new
      gem_pusher.handle_options [gem_path, '--host', artifactory_endpoint, '--key', 'artifactory_api_key', '--verbose']
      gem_pusher.execute
    end
  end
end
